from logging import Logger, StreamHandler
from pathlib import Path
from typing import ContextManager, Callable


def test_check_app_structure(local, app_dir: Path,
                             catch_log_warnings: Callable[[Logger], ContextManager[StreamHandler]]):
    from fastgenomics import app_checker

    with catch_log_warnings(app_checker.logger) as handler:
        app_checker.check_app_structure(app_dir)
        w1, w2, w3 = handler.stream.getvalue().strip().split('\n')

    assert 'LICENSE is missing!' in str(w1)
    assert 'requirements.txt is missing!' in str(w2)
    assert 'No sample_data' in str(w3)



# Flit documentation build configuration file
# http://www.sphinx-doc.org/en/stable/config.html

import sys
from pathlib import Path
sys.path.insert(0, str(Path(__file__).parent.parent))
from fastgenomics import __version__


needs_sphinx = '1.3'

extensions = [
    'sphinx.ext.autodoc',      # API docs
    'sphinx.ext.napoleon',     # google-style docstrings
    'sphinx.ext.intersphinx',  # links to official docs
    'sphinx_autodoc_napoleon_typehints',
]

intersphinx_mapping = dict(
    python=('https://docs.python.org/3', None),
    jsonschema=('https://python-jsonschema.readthedocs.io/en/v2.6.0/', None),
)

templates_path = ['_templates']
exclude_patterns = ['_build']
html_static_path = ['_static']

source_suffix = '.rst'
master_doc = 'index'

project = u'FASTGenomics'
copyright = u'2018, Comma Soft'

release = __version__
version = __version__  # short version

pygments_style = 'sphinx'
html_theme = 'sphinx_rtd_theme'
#html_theme_options = {}

#html_logo = '_static/flit_logo_nobg_cropped.svg'
#html_favicon = None

html_context = {
    'display_github': True,
    'github_user': 'fastgenomics',
    'github_repo': 'fastgenomics-py',
    'github_version': 'master',
    'conf_py_path': '/doc/',
}




import pytest

from io import StringIO
from logging import StreamHandler, WARNING, Logger
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, ContextManager

HERE = Path(__file__).parent
APP_DIR = HERE / 'sample_app'
DATA_ROOT = HERE / 'sample_data'


@pytest.fixture
def app_dir():
    return APP_DIR


@pytest.fixture
def data_root():
    return DATA_ROOT


@pytest.fixture
def catch_log_warnings() -> Callable[[Logger], ContextManager[StreamHandler]]:
    @contextmanager
    def catcher(logger: Logger):
        handler = StreamHandler(StringIO())
        handler.setLevel(WARNING)
        logger.addHandler(handler)
        yield handler

    return catcher


def get_local_paths():
    return dict(
        app=APP_DIR,
        data=DATA_ROOT / 'data',
        config=DATA_ROOT / 'config',
        summary=DATA_ROOT / 'summary',
        output=DATA_ROOT / 'output',
    )


@pytest.fixture
def local(monkeypatch):
    """patches the paths for local testing"""
    monkeypatch.setattr("fastgenomics._common.DEFAULT_APP_DIR", str(APP_DIR))
    monkeypatch.setattr("fastgenomics._common.DEFAULT_DATA_ROOT", str(DATA_ROOT))
    monkeypatch.setattr("fastgenomics._common._PATHS", {})
    monkeypatch.setattr("fastgenomics.io._PARAMETERS", {})
    monkeypatch.setattr("fastgenomics._common._MANIFEST", {})
    monkeypatch.setattr("fastgenomics._common._INPUT_FILE_MAPPING", {})


@pytest.fixture
def fg_env(monkeypatch):
    """sets app_dir and data_root by env-variables"""
    monkeypatch.setenv('FG_APP_DIR', str(APP_DIR))
    monkeypatch.setenv('FG_DATA_ROOT', str(DATA_ROOT))
    monkeypatch.setattr("fastgenomics._common._PATHS", {})
    monkeypatch.setattr("fastgenomics.io._PARAMETERS", {})
    monkeypatch.setattr("fastgenomics._common._MANIFEST", {})
    monkeypatch.setattr("fastgenomics._common._INPUT_FILE_MAPPING", {})


@pytest.fixture
def clear_output():
    """clear everything except of .gitignore"""
    for name in ['output', 'summary']:
        sub_dir = DATA_ROOT / name
        for entry in sub_dir.glob('*.*'):
            if entry.name != '.gitignore':
                entry.unlink()


@pytest.fixture
def fake_docker(monkeypatch):
    """fakes the docker-environment by overriding the running_within_docker-method returning always true"""
    monkeypatch.setattr("fastgenomics.tools.running_within_docker", lambda: True)



"""
FASTGenomics tools for app development
"""
from pathlib import Path
from logging import getLogger


logger = getLogger('fastgenomics.tools')


def running_within_docker() -> bool:
    """
    detects, if module is running within docker and returns the result as bool
    """
    if Path('/.dockerenv').exists():
        logger.debug("Running within docker")
        return True
    else:
        logger.info("Running locally")
        return False



"""
FASTGenomics IO helpers
=======================

Wraps input/output of ``input_file_mapping.json`` and parameters given by the FASTGenomics runtime.

If you want to work without docker, you can set two environment variables to ease testing:

``APP_ROOT_DIR``: This path should contain manifest.json, normally this is /app.
``DATA_ROOT_DIR``: This path should contain you test data - normally, this is /fastgenomics.

You can set them by environment variables or just call ``fg_io.set_paths(path_to_app, path_to_data_root)``
"""
import os
import typing as ty
from pathlib import Path
from logging import getLogger

from . import _common, tools


logger = getLogger('fastgenomics.io')


_PARAMETERS = {}


class UnknownAppTypeError(Exception):
    """
    Exception for unsupported App types.

    Args:
        typ: FASTGenomics App type
    """
    def __init__(self, typ: str):
        self.typ = typ

    @classmethod
    def check_type(cls, typ: str):
        if typ != 'Calculation':
            raise cls(typ)

    @property
    def message(self):
        return f'File output for “{self.typ}” applications not supported!'

    def __str__(self):
        return f'Not Supported: {self.message}'

    def __repr__(self):
        return f'{self.__class__.__name__}({self.typ!r})'


def set_paths(
    app_dir: ty.Optional[str] = None,
    data_root: ty.Optional[str] = None,
):
    """
    Sets the paths for the module to search for the ``manifest.json`` and IO/files

    The following strategy is used:

    - use variables provided to function, but warn if running within docker
    - else: if set, use environment-variables ``FG_APP_DIR`` and ``FG_DATA_ROOT``
    - else: if running within docker use defaults ``/app`` and ``/fastgenomics``
    - else: raise exception

    Args:
        app_dir: Path to root directory of the application (must contain ``manifest.json``)
        data_root: Path to root directory of input/output/config/summary
    
    Raises:
        FileNotFoundError: If any mapped paths do not point to files.
    """
    if tools.running_within_docker() is True:
        if any([app_dir, data_root, os.environ.get("FG_APP_DIR"), os.environ.get("FG_DATA_ROOT")]):
            logger.warning("Running within docker - non-default paths may result in errors!")

    if app_dir is None:
        app_dir_path = Path(os.environ.get("FG_APP_DIR", _common.DEFAULT_APP_DIR)).absolute()
    else:
        app_dir_path = Path(app_dir).absolute()

    if data_root is None:
        data_root_path = Path(os.environ.get("FG_DATA_ROOT", _common.DEFAULT_DATA_ROOT)).absolute()
    else:
        data_root_path = Path(data_root).absolute()

    logger.info(f"Using {app_dir_path} as app directory")
    logger.info(f"Using {data_root_path} as data root")

    # set paths
    paths = dict(
        app=app_dir_path,
        data=data_root_path / Path('data'),
        config=data_root_path / Path('config'),
        output=data_root_path / Path('output'),
        summary=data_root_path / Path('summary'),
    )

    # check and set
    _common.check_paths(paths)
    _common._PATHS = paths


def get_input_path(input_key: str) -> Path:
    """
    Retrieves the location of a input file.
    Keep in mind that you have to define your input files in your ``manifest.json`` in advance!

    Args:
        input_key: Key of an input file

    Returns:
        Location of the input file

    Raises:
        KeyError: If the ``input_key`` is not defined in ``manifest.json``
        FileNotFoundError: If the mapped path does not point to a file
    """
    manifest = _common.get_app_manifest()['Input']
    input_file_mapping = _common.get_input_file_mapping()

    # check for key in manifest
    if input_key not in manifest:
        err_msg = f"Input '{input_key}' not defined in manifest.json!"
        logger.error(err_msg)
        raise KeyError(err_msg)

    # check for key in mapping
    if input_key not in input_file_mapping:
        err_msg = f"Input '{input_key}' not defined in input_file_mapping!"
        logger.error(err_msg)
        raise KeyError(err_msg)

    # check existence
    input_file = input_file_mapping[input_key]
    if not input_file.exists():
        err_msg = f"Input-file '{input_file}' not found! Please check your input_file_mapping."
        logger.error(err_msg)
        raise FileNotFoundError(err_msg)

    return input_file


def get_output_path(output_key: str) -> Path:
    """
    Retrieves the location of the output file.
    Keep in mind that you have to define your output files in your ``manifest.json`` in advance!

    You can use this path-object to write your output as follows::

        my_path_object = get_output_path('my_output_key')
        with my_path_object.open('w', encoding='utf-8') as f_out:
            f_out.write("something")

    Args:
        output_key: Key of an output file

    Returns:
        Location of the output file
    
    Raises:
        UnknownAppTypeError: If the app type is unknown
        KeyError: If the ``output_key`` is not defined
    """
    manifest = _common.get_app_manifest()

    UnknownAppTypeError.check_type(manifest['Type'])

    # get output_file_mapping
    output_file_mapping = manifest['Output']
    if output_key not in output_file_mapping:
        err_msg = f"Key '{output_key}' not defined in manifest.json!"
        logger.error(err_msg)
        raise KeyError(err_msg)

    output_file = _common.get_paths()['output'] / output_file_mapping[output_key]['FileName']

    # check for existence
    if output_file.exists():
        err_msg = f"Output-file '{output_file}' already exists!"
        logger.warning(err_msg)

    return output_file


def get_summary_path() -> Path:
    """
    Retrieves the location of the summary file.
    Please write your summary as CommonMark-compatible Markdown into this file.

    Returns:
        Location of the summary file
    
    Raises:
        UnknownAppTypeError: If the app type is unknown
    """
    manifest = _common.get_app_manifest()

    UnknownAppTypeError.check_type(manifest['Type'])

    output_file = _common.get_paths()['summary'] / 'summary.md'

    # check for existence
    if output_file.exists():
        err_msg = f"Summary-file '{output_file}' already exists!"
        logger.warning(err_msg)

    return output_file


def get_parameters() -> _common.Parameters:
    """
    Returns a dict of all parameters along with its current value provided by
    ``parameters.json``, falling back to defaults defined in ``manifest.json``.

    Unset optional Parameters result in ``None`` values.

    Returns:
        Parameter dict with current values
    """
    global _PARAMETERS

    # use cache and return dict of {param: current value}
    if _PARAMETERS:
        return _return_parameters()

    # else: load parameters
    parameters = _common.load_parameters_from_manifest()
    runtime_parameters = _common.load_runtime_parameters()

    # merge with defaults
    for name, current_value in runtime_parameters.items():
        if name not in parameters:
            logger.warning(f"Ignoring runtime parameter {name}, as it is not defined in manifest.json!")
            continue
        parameters[name] = _update_param_value(parameters[name], current_value)

    # check types
    _common.check_parameter_types(parameters)

    # update cache and return
    _PARAMETERS = parameters
    return _return_parameters()


def _return_parameters() -> _common.Parameters:
    """helper for structured return of parameters"""
    return {name: param.value for name, param in _PARAMETERS.items()}


def _update_param_value(param: _common.Parameter, new_value: ty.Any):
    """helper function for updating the value of a Parameter instance"""
    return _common.Parameter(
        name=param.name,
        type=param.type,
        value=new_value,
        default=param.default,
        optional=param.optional,
        enum=param.enum,
        description=param.description,
    )


def get_parameter(param_key: str) -> _common.ParameterValue:
    """
    Get a specific parameter’s current value or the default.
    Provide a key defined in the ``manifest.json``.
    
    Unset optional Parameters result in ``None`` being returned.
    
    Args:
        param_key: The parameter’s key
    
    Returns:
        The parameter’s value.
    
    Raises:
        KeyError: If a parameter named ``param_key`` is not defined
    """
    parameters = get_parameters()

    # check for existence and return or raise exception
    if param_key not in parameters:
        raise KeyError(f"Parameter {param_key} not defined in manifest.json!")
    return parameters[param_key]



import functools

import pkg_resources
import jinja2

from . import __name__ as module_name


j2env = jinja2.Environment(loader=jinja2.PackageLoader(module_name, 'templates'))

resource_stream = functools.partial(pkg_resources.resource_stream, module_name)



"""
FASTGenomics Test-Suite
=======================

Provides methods to check your app structure, ``manifest.json`` and ``input_file_mapping.json``.
"""
from pathlib import Path
from logging import getLogger

from . import _common, io


logger = getLogger('fastgenomics.testing')


def check_app_structure(app_dir: Path):
    """
    Checks the structure of your app –
    only checks for mandatory files and directories.

    It only logs warnings if default parameter types are off.
    
    Args:
        app_dir: Root directory of the app

    Raises:
        AssertionError: If any vital files are missing
        ~jsonschema.exceptions.ValidationError: If the manifest isn’t valid
    """

    # check app structure
    logger.info(f"Checking app-structure in {app_dir}")
    to_check = ['manifest.json', 'README.md', 'LICENSE', 'Dockerfile', 'requirements.txt']
    warn_only = ['LICENSE', 'requirements.txt']

    missing = []
    for entry in to_check:
        entry_path = app_dir / entry
        if not entry_path.exists():
            err_msg = f"{entry_path} is missing!"
            if entry in warn_only:
                logger.warning(err_msg)
            else:
                logger.error(err_msg)
                missing.append(entry_path)
    assert not missing, missing

    # check manifest.json
    logger.info(f"Checking manifest.json in {app_dir}")
    manifest = _common.get_app_manifest()
    # This is already done in get_app_manifest, but let’s make sure this is tested
    _common.assert_manifest_is_valid(dict(FASTGenomicsApplication=manifest))

    # checking for sample_data
    logger.info(f"Checking for sample_data in {app_dir}")
    sample_dir = app_dir / 'sample_data'

    valid_sample_data_sub_dirs = ['data', 'config']
    if manifest['Type'] == 'Calculation':
        valid_sample_data_sub_dirs += ['output', 'summary']

    if not sample_dir.exists():
        logger.warning("No sample_data found - please provide sample data!")
    else:
        for sub_dir in valid_sample_data_sub_dirs:
            assert (sample_dir / sub_dir).exists(), f"sample_data subdirectory {sub_dir} is missing!"


def check_input_file_mapping(app_dir: Path):
    """
    Checks the input_file_mapping
    
    Args:
        app_dir: Root directory of the app

    Raises:
        FileNotFoundError: If any mapped paths do not point to files.
        ~json.JSONDecodeError: If the manifest is not valid JSON
        KeyError: If an file in the manifest is not mapped
    """
    sample_dir = app_dir / 'sample_data'
    io.set_paths(str(app_dir), str(sample_dir))
    _common.get_input_file_mapping(check_mapping=True)



"""
FASTGenomics App Creation Suite
===============================

Provides methods to create basic boilerplate for your apps.
"""
from pathlib import Path
from logging import getLogger

from ._resources import j2env
from . import _common


logger = getLogger('fastgenomics.testing')


# registry
DOCKER_REGISTRY = 'apps.fastgenomics.org'


def create_docker_compose(app_dir: Path, app_name: Path, sample_dir: Path, docker_registry: str = DOCKER_REGISTRY):
    """
    Creates an ``docker-compose.test.yml`` for testing
    
    Args:
        app_dir: Root directory of the app
        app_name: Name of the app, ``snake_case``
        sample_dir: Data root dir containing ``data/**`` and ``config/input_file_mapping.json``
        docker_registry: App registry
    """
    docker_compose_file = app_dir / 'docker-compose.test.yml'

    if docker_compose_file.exists():
        logger.warning(f"{docker_compose_file.name} already existing! Aborting.")
        return

    # get app type
    manifest = _common.get_app_manifest()
    app_type = manifest['Type']

    logger.info("Loading docker-compose.test.yml template")
    template = j2env.get_template('docker-compose.yml.j2')

    logger.info(f"Writing {docker_compose_file}")
    with docker_compose_file.open('w') as f_out:
        temp = template.render(
            app_name=app_name,
            sample_dir=sample_dir.relative_to(app_dir),
            docker_registry=docker_registry,
            app_type=app_type,
        )
        f_out.write(temp)


def create_file_mapping(sample_dir: Path):
    """
    Creates a base ``input_file_mapping.json``
    
    Args:
        sample_dir: Sample data directory that will contain ``data/**`` and ``config/input_file_mapping.json``
    """
    sample_output_dir = sample_dir / 'data' / 'other_app_uuid' / 'output'
    file_mapping_file = sample_dir / 'config' / 'input_file_mapping.json'

    if file_mapping_file.exists():
        logger.warning(f"{file_mapping_file} already existing! Aborting.")
        return

    # creating output directories
    sample_output_dir.mkdir(parents=True, exist_ok=True)
    file_mapping_file.parent.mkdir(parents=True, exist_ok=True)

    # create file_mappings
    manifest = _common.get_app_manifest()
    input_keys = manifest['Input'].keys()
    file_mapping = {key: sample_output_dir / 'fix_me.txt' for key in input_keys}

    # write file_mappings
    logger.info("Loading input_file_mapping.json template")
    template = j2env.get_template('input_file_mapping.json.j2')

    logger.info(f"Writing {file_mapping_file}")
    with file_mapping_file.open('w') as f_out:
        temp = template.render(file_mapping=file_mapping)
        f_out.write(temp)

    print()
    print(f"Please edit {file_mapping_file} and provide the following files:")
    for key in input_keys:
        print(f" - {key}: {manifest['Input'][key]['Usage']} ({manifest['Input'][key]['Type']})")
    print()



"""
FASTGenomics common IO helpers: Wraps manifest.json, input_file_mapping.json and parameters.json given by the
FASTGenomics runtime.

If you want to work without docker, you can set two environment variables to ease testing:

``APP_ROOT_DIR``: This path should contain manifest.json, normally this is /app.
``DATA_ROOT_DIR``: This path should contain you test data - normally, this is /fastgenomics.

You can set them by environment variables or just call ``fastgenomics.common.set_paths(path_to_app, path_to_data_root)``
"""
import os
import json
import typing as ty
from pathlib import Path
from logging import getLogger

import jsonschema

from ._resources import resource_stream


logger = getLogger('fastgenomics.common')

# set default paths
DEFAULT_APP_DIR = '/app'
DEFAULT_DATA_ROOT = '/fastgenomics'

# init cache
_PATHS = {}
_MANIFEST = {}
_INPUT_FILE_MAPPING = {}


ParameterValue = ty.Union[dict, list, bool, int, float, str, None]
Parameters = ty.Dict[str, ParameterValue]
PathsDict = ty.Dict[str, Path]
FileMapping = ty.Dict[str, Path]


class Parameter(ty.NamedTuple):
    """Parameter entry"""
    name: str
    type: str
    value: ty.Any  # uses default for initialization or None
    default: ty.Any
    optional: bool
    enum: ty.Optional[ty.List[ty.Any]]
    description: str


def check_paths(paths: PathsDict, raise_error: bool = True):
    """
    checks, if main paths are existing along with the manifest.json and input_file_mapping.json
    """
    # check base paths
    for dir_to_check in ['app', 'config', 'data']:
        path_entry = paths.get(dir_to_check)
        if path_entry is None or not path_entry.exists():
            err_msg = f"Path to {path_entry} not found! Check paths!"
            if raise_error is True:
                raise FileNotFoundError(err_msg)
            else:
                logger.warning(err_msg)

    # check configs files
    paths_to_check = [paths['app'] / 'manifest.json']
    if not os.environ.get('INPUT_FILE_MAPPING'):
        paths_to_check.append(paths['config'] / 'input_file_mapping.json')

    for path_entry in paths_to_check:
        if not path_entry.exists():
            err_msg = f"`{path_entry}` does not exist! Please check paths and existence!"
            if raise_error is True:
                raise FileNotFoundError(err_msg)
            else:
                logger.warning(err_msg)


def check_input_file_mapping(input_file_mapping: FileMapping):
    """checks the keys in input_file_mapping and existence of the files

    raises a KeyError on missing Key and FileNotFoundError on missing file
    """
    manifest = get_app_manifest()['Input']

    not_in_manifest = set(input_file_mapping.keys()) - set(manifest.keys())
    not_in_ifm = set(manifest.keys()) - set(input_file_mapping.keys())

    optional = set()  # not implemented yet
    missing = not_in_ifm - optional

    # check keys
    if not_in_manifest:
        logger.warning(f"Ignoring Keys defined in input_file_mapping: {not_in_manifest}")

    if optional:   # not implemented yet
        logger.warning(f"Non-optional keys not defined in input_file_mapping: {missing}")

    if missing:
        raise KeyError(f"Non-optional keys not defined in input_file_mapping: {missing}")

    # check for existence
    for entry in input_file_mapping.values():
        if not entry.exists():
            raise FileNotFoundError(f"{entry}, defined in input_file_mapping, not found!")


def str_to_path_file_mapping(relative_mapping: ty.Dict[str, str]) -> FileMapping:
    """maps the relative string paths given in input_file_mapping to absolute paths"""
    data_path = get_paths()['data']

    absolute_mapping = {key: data_path / mapped_file
                        for key, mapped_file in relative_mapping.items()}
    return absolute_mapping


def load_input_file_mapping() -> ty.Dict[str, str]:
    """helper function loading the input_file_mapping either from environment or from file"""
    # try to get input file mapping from environment
    empty_str = '{}'
    ifm_str = os.environ.get('INPUT_FILE_MAPPING', empty_str)
    source_str = "`INPUT_FILE_MAPPING` environment"

    # else use input_file_mapping file
    if ifm_str == empty_str:
        ifm_path = get_paths()['config'] / 'input_file_mapping.json'
        if not ifm_path.exists():
            raise FileNotFoundError(f"Input file mapping {ifm_path} not found!")

        with open(get_paths()['config'] / 'input_file_mapping.json', encoding='utf-8') as f:
            ifm_str = f.read()
            source_str = ifm_path.name
    logger.info(f"Input file mapping loaded from {source_str}.")

    # decode json:
    try:
        ifm_dict = json.loads(ifm_str)
    except json.JSONDecodeError as e:
        e.msg = f"{source_str} is not valid JSON: {e.msg}"
        raise e
    return ifm_dict


def get_input_file_mapping(check_mapping: bool = True) -> FileMapping:
    """returns the input_file_mapping either from environment `INPUT_FILE_MAPPING` or from config file"""
    global _INPUT_FILE_MAPPING

    if _INPUT_FILE_MAPPING:
        return _INPUT_FILE_MAPPING

    # load mapping
    ifm_dict = load_input_file_mapping()

    # convert into paths
    input_file_mapping = str_to_path_file_mapping(ifm_dict)

    # check existence
    if check_mapping:
        check_input_file_mapping(input_file_mapping)

    # update cache and return
    _INPUT_FILE_MAPPING = input_file_mapping
    return _INPUT_FILE_MAPPING


def get_paths() -> PathsDict:
    """
    safe getter for the runtime paths

    if paths are not initialized, it runs ``set_paths(DEFAULT_APP_DIR, DEFAULT_DATA_ROOT)``
    """
    if not _PATHS:
        from . import io
        io.set_paths()
    return _PATHS


def assert_manifest_is_valid(config: dict):
    """
    Asserts that the manifest (``manifest.json``) matches our JSON-Schema.
    If not a :py:exc:`jsonschema.ValidationError` will be raised.
    """
    
    with resource_stream('schemes/manifest_schema.json') as f:
        schema = json.load(f)
    jsonschema.validate(config, schema)

    parameters = config["FASTGenomicsApplication"]["Parameters"]
    if parameters is not None:
        for name, properties in parameters.items():
            expected_type = properties["Type"]
            enum = properties.get("Enum")
            default_value = properties["Default"]
            optional = properties.get("Optional", False)
            warn_if_not_of_type(name, expected_type, enum, default_value, optional, is_default=True)


def get_app_manifest() -> dict:
    """
    Parses and returns the app manifest.json

    Raises:
        RuntimeError: If ``manifest.json`` does not exist.
    """
    global _MANIFEST

    # use cache
    if _MANIFEST:
        return _MANIFEST

    manifest_file = get_paths()['app'] / 'manifest.json'
    if not manifest_file.exists():
        err_msg = (f"App manifest {manifest_file} not found! "
                   "Please provide a manifest.json in the application's root-directory.")
        raise RuntimeError(err_msg)
    with open(manifest_file, encoding='utf-8') as f:
        try:
            config = json.load(f)
            assert_manifest_is_valid(config)
            # update cache
            _MANIFEST = config['FASTGenomicsApplication']
        except json.JSONDecodeError:
            err_msg = f"App manifest {manifest_file} not a valid JSON-file - check syntax!"
            raise RuntimeError(err_msg)

    return _MANIFEST


def _update_param_value(param: Parameter, new_value: ty.Any):
    """helper function for updating the value of a Parameter instance"""
    return Parameter(name=param.name,
                     type=param.type,
                     value=new_value,
                     default=param.default,
                     optional=param.optional,
                     enum=param.enum,
                     description=param.description)


def load_runtime_parameters() -> Parameters:
    """loads and returns the runtime parameters from parameters.json"""

    parameters_file = get_paths()['config'] / 'parameters.json'

    if not parameters_file.exists():
        logger.info(f"No runtime parameters {parameters_file} found - using defaults.")
        return {}

    try:
        runtime_parameters = json.loads(parameters_file.read_text())
    except json.JSONDecodeError:
        logger.error(
            f"Could not read {parameters_file} due to an unexpected error. "
            f"Please report this at https://github.com/FASTGenomics/fastgenomics-py/issues")
        return {}

    return runtime_parameters


def load_parameters_from_manifest() -> ty.Dict[str, Parameter]:
    """returns the parameters section defined in the manifest.json as dict of Parameter
    """
    param_section = get_app_manifest()['Parameters']

    return {name: Parameter(name=name,
                            type=value['Type'],
                            value=value.get('Default'),
                            default=value.get('Default'),
                            optional=value.get('Optional', False),
                            enum=value.get('Enum'),
                            description=value['Description'])

            for name, value in param_section.items()}


def value_is_of_type(expected_type: str, enum: ty.Optional[list], value: ty.Any, optional: bool) -> bool:
    """tests, of a value is an instance of a given an expected type"""
    type_mapping = {
        'float': (int, float),
        'integer': int,
        'bool': bool,
        'list': list,
        'dict': dict,
        'string': str,
        'enum': object,
    }
    mapped_type = type_mapping.get(expected_type)
    if mapped_type is None:
        raise ValueError(f"Unknown type to check: {expected_type}")

    if optional and value is None:
        return True
    if enum is not None:
        if expected_type != 'enum':
            raise ValueError(f"Enum provided but type is {expected_type}")
        return value in enum
    return isinstance(value, mapped_type)


def check_parameter_types(parameters: Parameters):
    """checks the correct type of parameters as specified in the manifest.json"""
    manifest_parameters = load_parameters_from_manifest()

    for param_name, param in parameters.items():
        manifest_param = manifest_parameters[param_name]
        # parameter types see manifest_schema.json
        # we do not throw an exception because having multi-value parameters is
        #  common in some libraries, e.g. specify "red" or 24342
        warn_if_not_of_type(name=param_name, expected_type=manifest_param.type, enum=manifest_param.enum,
                            value=param.value, optional=manifest_param.optional)


def warn_if_not_of_type(name, expected_type, enum, value, optional, is_default=False):
    if value_is_of_type(expected_type, enum, value, optional):
        return

    msg = f"The {'default ' if is_default else ''}parameter {name} has a different value than expected. "
    if enum is None:
        msg += f"It should be a {expected_type} but is a {type(value)}. "
    else:
        msg += f"It should be one of {enum!r} but is {value!r}. "
    logger.warning(msg + f"The value is accessible but beware!")



import pytest
import pathlib

from fastgenomics import _common, io


def test_paths_are_initialized(local):
    _common.get_paths()


def test_custom_init_paths(app_dir, data_root):
    io.set_paths(app_dir, data_root)
    _common.get_paths()


def test_paths_from_env(fg_env):
    _common.get_paths()


def test_cannot_init_nonexisting_paths():
    with pytest.raises(FileNotFoundError):
        io.set_paths("i_don't_exist", "me_neither")


def test_custom_init_path_within_docker(fake_docker, app_dir, data_root):
    with pytest.warns(None):
        io.set_paths(app_dir, data_root)
        _common.get_paths()


def test_get_app_manifest(local):
    _common.get_app_manifest()


def test_assert_manifest_is_valid(local):
    manifest = _common.get_app_manifest()
    _common.assert_manifest_is_valid({'FASTGenomicsApplication': manifest})


def test_can_get_parameters(local):
    parameters = io.get_parameters()
    assert len(parameters) > 0


def test_get_paramerets_dont_run_into_recursion(monkeypatch, local):
    monkeypatch.setattr("fastgenomics._common.load_parameters_from_manifest", dict)
    monkeypatch.setattr("fastgenomics._common.load_runtime_parameters", dict)
    io.get_parameters()


def test_parameters(local):
    parameters = io.get_parameters()

    assert "StrValue" in parameters
    assert parameters["StrValue"] == "hello from parameters.json"

    assert "IntValue" in parameters
    assert parameters["IntValue"] == 150

    assert "FloatValue" in parameters
    assert parameters["FloatValue"] == float(100)

    assert "BoolValue" in parameters
    assert parameters["BoolValue"] is True

    assert "ListValue" in parameters
    assert parameters["ListValue"] == [1, 2, 3]

    assert "DictValue" in parameters
    assert parameters["DictValue"] == {"foo": 42, "bar": "answer to everything"}

    assert "OptionalIntValueConcrete" in parameters
    assert parameters["OptionalIntValueConcrete"] == 4

    assert "OptionalIntValueNull" in parameters
    assert parameters["OptionalIntValueNull"] is None

    assert "EnumValue" in parameters
    assert parameters["EnumValue"] == "X"


def test_can_get_specific_parameter(local):
    assert io.get_parameter("IntValue") == 150


def test_can_get_null_parameter(local):
    assert io.get_parameter("OptionalIntValueNull") is None


def test_can_have_different_type(local, monkeypatch):
    # patch custom parameter load function
    monkeypatch.setattr("fastgenomics._common.load_runtime_parameters", lambda: {"StrValue": 1})

    # get parameters and compare parameters of different types
    with pytest.warns(None):
        parameters = io.get_parameters()
        assert 1 == parameters["StrValue"]


def test_load_input_file_mapping(local):
    input_file_mapping = _common.load_input_file_mapping()
    assert "some_input" in input_file_mapping


def test_input_file_mapping_to_paths(local):
    ifm_dict = _common.load_input_file_mapping()
    input_file_mapping = _common.str_to_path_file_mapping(ifm_dict)
    assert isinstance(input_file_mapping['some_input'], pathlib.Path)
    assert input_file_mapping['some_input'].exists()


def test_check_input_file_mapping(local):
    ifm_dict = _common.load_input_file_mapping()
    input_file_mapping = _common.str_to_path_file_mapping(ifm_dict)

    # test everything is ok
    _common.check_input_file_mapping(input_file_mapping)

    # test if additional keys trigger warning
    input_file_mapping['unused_key'] = pathlib.Path(".")
    with pytest.warns(None):
        _common.check_input_file_mapping(input_file_mapping)

    # test raises KeyError on missing entry
    with pytest.raises(KeyError):
        _common.check_input_file_mapping({})

    # test raises FileNotFoundError on wrong item
    with pytest.raises(FileNotFoundError):
        _common.check_input_file_mapping({"some_input": pathlib.Path("i_don't_exist")})


def test_load_input_file_mapping_from_env(local, monkeypatch):
    monkeypatch.setenv('INPUT_FILE_MAPPING', '{"some_key_from_env": "some_value"}')
    monkeypatch.setattr('pathlib.Path.exists', lambda self: False if "input_file_mapping.json" in self.name else True)
    input_file_mapping = _common.load_input_file_mapping()
    assert "some_input" not in input_file_mapping, "input_file_mapping from file used instead if env!"
    assert "some_key_from_env" in input_file_mapping
    # test if rest of _common still works
    _ = _common.get_app_manifest()
    _ = io.get_parameters()


def test_get_input_file_mapping(local):
    input_file_mapping = _common.get_input_file_mapping()
    assert "some_input" in input_file_mapping
    assert input_file_mapping['some_input'].exists()


def test_get_input_file_mapping_from_env(local, monkeypatch):
    monkeypatch.setenv('INPUT_FILE_MAPPING', '{"some_input": "input.csv", "some_key_from_env": "input.csv"}')
    input_file_mapping = _common.get_input_file_mapping()
    assert input_file_mapping['some_key_from_env'].exists()



import fastgenomics.io as fg_io
import pytest


def test_can_read_input_file(local):
    # can get path
    to_test = fg_io.get_input_path("some_input")

    # path exists
    assert to_test.exists()


def test_cannot_read_undefined_input(local):
    with pytest.raises(KeyError):
        fg_io.get_input_path("i_don't_exist")


def test_can_write_summary(local, clear_output):
    sum_file = fg_io.get_summary_path()
    with sum_file.open('w', encoding='utf-8') as out:
        out.write('test')
    assert sum_file.exists()


def test_can_write_output(local, clear_output):
        out_path = fg_io.get_output_path("some_output")
        assert out_path.name == 'some_output.csv'
        with out_path.open('w', encoding='utf-8') as out:
            out.write('test')
        assert out_path.exists()


def test_cannot_write_undefined_output(local):
    with pytest.raises(KeyError):
        fg_io.get_output_path("i_don't_exist")


# test things, imported from fastgenomics.common, are available
def test_import_from_common(app_dir, data_root):
    fg_io.set_paths(app_dir=app_dir, data_root=data_root)
    assert len(fg_io.get_parameters()) > 0
    fg_io.get_parameter('IntValue')




#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastgenomics']

package_data = \
{'': ['*'], 'fastgenomics': ['schemes/*', 'templates/*']}

install_requires = \
['jsonschema~=2.6', 'Jinja2~=2.9']

setup(name='fastgenomics',
      version='0.6.3',
      description='FASTGenomics python helper',
      author='FASTGenomics Team',
      author_email='contact@fastgenomics.org',
      url='https://github.com/fastgenomics/fastgenomics-py',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='~=3.6',
     )
